open Base
open Hardcaml
open Hardcaml_waveterm
open Hardcaml_circuits

let init ~cycles num_data =
  let open Signal in
  let clock = input "clock" 1 in
  let clear = input "clear" 1 in
  let data =
    let input_n name index width = input (name ^ Int.to_string index) width in
    List.init num_data ~f:(fun i ->
      { With_valid.valid = input_n "valid" (i + 1) 1; value = of_int ~width:8 (i + 1) })
  in
  let { With_valid.valid; value } =
    Pipelined_tree_mux.pipelined_tree_priority_select
      ~trace_reductions:true
      ~cycles
      ~reg:(reg (Reg_spec.create ~clock ~clear ()))
      data
  in
  let valid = output "valid" valid in
  let value = output "value" value in
  let sim =
    Cyclesim.create
      (Circuit.create_exn ~name:"pipelined_priority_select" [ valid; value ])
  in
  let waves, sim = Waveform.create sim in
  let valid =
    Array.init num_data ~f:(fun i ->
      Cyclesim.in_port sim ("valid" ^ Int.to_string (i + 1)))
  in
  waves, sim, valid
;;

let strobe ~cycles num_data =
  let waves, sim, valid = init ~cycles num_data in
  Cyclesim.cycle sim;
  for i = 0 to num_data - 1 do
    valid.(i) := Bits.vdd;
    Cyclesim.cycle sim;
    valid.(i) := Bits.gnd
  done;
  for _ = 1 to cycles do
    Cyclesim.cycle sim
  done;
  Cyclesim.cycle sim;
  Waveform.print
    ~display_width:86
    ~display_height:(11 + (num_data * 2))
    ~wave_width:1
    waves
;;

let%expect_test "combinational" =
  strobe ~cycles:0 1;
  [%expect
    {|
    ┌Signals───────────┐┌Waves───────────────────────────────────────────────────────────┐
    │valid1            ││    ┌───┐                                                       │
    │                  ││────┘   └───                                                    │
    │valid             ││    ┌───┐                                                       │
    │                  ││────┘   └───                                                    │
    │                  ││────────────                                                    │
    │value             ││ 01                                                             │
    │                  ││────────────                                                    │
    │                  ││                                                                │
    │                  ││                                                                │
    │                  ││                                                                │
    │                  ││                                                                │
    └──────────────────┘└────────────────────────────────────────────────────────────────┘ |}];
  strobe ~cycles:0 3;
  [%expect
    {|
    ┌Signals───────────┐┌Waves───────────────────────────────────────────────────────────┐
    │valid1            ││    ┌───┐                                                       │
    │                  ││────┘   └───────────                                            │
    │valid2            ││        ┌───┐                                                   │
    │                  ││────────┘   └───────                                            │
    │valid3            ││            ┌───┐                                               │
    │                  ││────────────┘   └───                                            │
    │valid             ││    ┌───────────┐                                               │
    │                  ││────┘           └───                                            │
    │                  ││────┬───┬───┬───────                                            │
    │value             ││ 03 │01 │02 │03                                                 │
    │                  ││────┴───┴───┴───────                                            │
    │                  ││                                                                │
    │                  ││                                                                │
    │                  ││                                                                │
    │                  ││                                                                │
    └──────────────────┘└────────────────────────────────────────────────────────────────┘ |}]
;;

let%expect_test "single cycle" =
  strobe ~cycles:1 1;
  [%expect
    {|
    ((cycle 0) (reductions (1)))
    ┌Signals───────────┐┌Waves───────────────────────────────────────────────────────────┐
    │clock             ││┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ │
    │                  ││  └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─│
    │clear             ││                                                                │
    │                  ││────────────────                                                │
    │valid1            ││    ┌───┐                                                       │
    │                  ││────┘   └───────                                                │
    │valid             ││        ┌───┐                                                   │
    │                  ││────────┘   └───                                                │
    │                  ││────┬───────────                                                │
    │value             ││ 00 │01                                                         │
    │                  ││────┴───────────                                                │
    └──────────────────┘└────────────────────────────────────────────────────────────────┘ |}];
  strobe ~cycles:1 3;
  [%expect
    {|
    ((cycle 0) (reductions (3)))
    ┌Signals───────────┐┌Waves───────────────────────────────────────────────────────────┐
    │clock             ││┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ │
    │                  ││  └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─│
    │clear             ││                                                                │
    │                  ││────────────────────────                                        │
    │valid1            ││    ┌───┐                                                       │
    │                  ││────┘   └───────────────                                        │
    │valid2            ││        ┌───┐                                                   │
    │                  ││────────┘   └───────────                                        │
    │valid3            ││            ┌───┐                                               │
    │                  ││────────────┘   └───────                                        │
    │valid             ││        ┌───────────┐                                           │
    │                  ││────────┘           └───                                        │
    │                  ││────┬───┬───┬───┬───────                                        │
    │value             ││ 00 │03 │01 │02 │03                                             │
    │                  ││────┴───┴───┴───┴───────                                        │
    └──────────────────┘└────────────────────────────────────────────────────────────────┘ |}]
;;

let%expect_test "2 cycles" =
  strobe ~cycles:2 1;
  [%expect
    {|
    ((cycle 0) (reductions (1)))
    ((cycle 1) (reductions (1)))
    ┌Signals───────────┐┌Waves───────────────────────────────────────────────────────────┐
    │clock             ││┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ │
    │                  ││  └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─│
    │clear             ││                                                                │
    │                  ││────────────────────                                            │
    │valid1            ││    ┌───┐                                                       │
    │                  ││────┘   └───────────                                            │
    │valid             ││            ┌───┐                                               │
    │                  ││────────────┘   └───                                            │
    │                  ││────────┬───────────                                            │
    │value             ││ 00     │01                                                     │
    │                  ││────────┴───────────                                            │
    └──────────────────┘└────────────────────────────────────────────────────────────────┘ |}];
  strobe ~cycles:2 3;
  [%expect
    {|
    ((cycle 0) (reductions (2 1)))
    ((cycle 1) (reductions (2)))
    ┌Signals───────────┐┌Waves───────────────────────────────────────────────────────────┐
    │clock             ││┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ │
    │                  ││  └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─│
    │clear             ││                                                                │
    │                  ││────────────────────────────                                    │
    │valid1            ││    ┌───┐                                                       │
    │                  ││────┘   └───────────────────                                    │
    │valid2            ││        ┌───┐                                                   │
    │                  ││────────┘   └───────────────                                    │
    │valid3            ││            ┌───┐                                               │
    │                  ││────────────┘   └───────────                                    │
    │valid             ││            ┌───────────┐                                       │
    │                  ││────────────┘           └───                                    │
    │                  ││────────┬───┬───┬───┬───────                                    │
    │value             ││ 00     │03 │01 │02 │03                                         │
    │                  ││────────┴───┴───┴───┴───────                                    │
    └──────────────────┘└────────────────────────────────────────────────────────────────┘ |}];
  strobe ~cycles:2 13;
  [%expect
    {|
    ((cycle 0) (reductions (4 4 4 1)))
    ((cycle 1) (reductions (4)))
    ┌Signals───────────┐┌Waves───────────────────────────────────────────────────────────┐
    │clock             ││┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ │
    │                  ││  └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─│
    │clear             ││                                                                │
    │                  ││────────────────────────────────────────────────────────────────│
    │valid1            ││    ┌───┐                                                       │
    │                  ││────┘   └───────────────────────────────────────────────────────│
    │valid10           ││                                        ┌───┐                   │
    │                  ││────────────────────────────────────────┘   └───────────────────│
    │valid11           ││                                            ┌───┐               │
    │                  ││────────────────────────────────────────────┘   └───────────────│
    │valid12           ││                                                ┌───┐           │
    │                  ││────────────────────────────────────────────────┘   └───────────│
    │valid13           ││                                                    ┌───┐       │
    │                  ││────────────────────────────────────────────────────┘   └───────│
    │valid2            ││        ┌───┐                                                   │
    │                  ││────────┘   └───────────────────────────────────────────────────│
    │valid3            ││            ┌───┐                                               │
    │                  ││────────────┘   └───────────────────────────────────────────────│
    │valid4            ││                ┌───┐                                           │
    │                  ││────────────────┘   └───────────────────────────────────────────│
    │valid5            ││                    ┌───┐                                       │
    │                  ││────────────────────┘   └───────────────────────────────────────│
    │valid6            ││                        ┌───┐                                   │
    │                  ││────────────────────────┘   └───────────────────────────────────│
    │valid7            ││                            ┌───┐                               │
    │                  ││────────────────────────────┘   └───────────────────────────────│
    │valid8            ││                                ┌───┐                           │
    │                  ││────────────────────────────────┘   └───────────────────────────│
    │valid9            ││                                    ┌───┐                       │
    │                  ││────────────────────────────────────┘   └───────────────────────│
    │valid             ││            ┌───────────────────────────────────────────────────│
    │                  ││────────────┘                                                   │
    │                  ││────────┬───┬───┬───┬───┬───┬───┬───┬───┬───┬───┬───┬───┬───┬───│
    │value             ││ 00     │0D │01 │02 │03 │04 │05 │06 │07 │08 │09 │0A │0B │0C │0D │
    │                  ││────────┴───┴───┴───┴───┴───┴───┴───┴───┴───┴───┴───┴───┴───┴───│
    └──────────────────┘└────────────────────────────────────────────────────────────────┘ |}]
;;

let%expect_test "more cycles" =
  strobe ~cycles:3 13;
  [%expect
    {|
    ((cycle 0) (reductions (3 3 3 3 1)))
    ((cycle 1) (reductions (3 2)))
    ((cycle 2) (reductions (2)))
    ┌Signals───────────┐┌Waves───────────────────────────────────────────────────────────┐
    │clock             ││┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ │
    │                  ││  └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─│
    │clear             ││                                                                │
    │                  ││────────────────────────────────────────────────────────────────│
    │valid1            ││    ┌───┐                                                       │
    │                  ││────┘   └───────────────────────────────────────────────────────│
    │valid10           ││                                        ┌───┐                   │
    │                  ││────────────────────────────────────────┘   └───────────────────│
    │valid11           ││                                            ┌───┐               │
    │                  ││────────────────────────────────────────────┘   └───────────────│
    │valid12           ││                                                ┌───┐           │
    │                  ││────────────────────────────────────────────────┘   └───────────│
    │valid13           ││                                                    ┌───┐       │
    │                  ││────────────────────────────────────────────────────┘   └───────│
    │valid2            ││        ┌───┐                                                   │
    │                  ││────────┘   └───────────────────────────────────────────────────│
    │valid3            ││            ┌───┐                                               │
    │                  ││────────────┘   └───────────────────────────────────────────────│
    │valid4            ││                ┌───┐                                           │
    │                  ││────────────────┘   └───────────────────────────────────────────│
    │valid5            ││                    ┌───┐                                       │
    │                  ││────────────────────┘   └───────────────────────────────────────│
    │valid6            ││                        ┌───┐                                   │
    │                  ││────────────────────────┘   └───────────────────────────────────│
    │valid7            ││                            ┌───┐                               │
    │                  ││────────────────────────────┘   └───────────────────────────────│
    │valid8            ││                                ┌───┐                           │
    │                  ││────────────────────────────────┘   └───────────────────────────│
    │valid9            ││                                    ┌───┐                       │
    │                  ││────────────────────────────────────┘   └───────────────────────│
    │valid             ││                ┌───────────────────────────────────────────────│
    │                  ││────────────────┘                                               │
    │                  ││────────────┬───┬───┬───┬───┬───┬───┬───┬───┬───┬───┬───┬───┬───│
    │value             ││ 00         │0D │01 │02 │03 │04 │05 │06 │07 │08 │09 │0A │0B │0C │
    │                  ││────────────┴───┴───┴───┴───┴───┴───┴───┴───┴───┴───┴───┴───┴───│
    └──────────────────┘└────────────────────────────────────────────────────────────────┘ |}];
  strobe ~cycles:4 13;
  [%expect
    {|
    ((cycle 0) (reductions (2 2 2 2 2 2 1)))
    ((cycle 1) (reductions (2 2 2 1)))
    ((cycle 2) (reductions (2 2)))
    ((cycle 3) (reductions (2)))
    ┌Signals───────────┐┌Waves───────────────────────────────────────────────────────────┐
    │clock             ││┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ │
    │                  ││  └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─│
    │clear             ││                                                                │
    │                  ││────────────────────────────────────────────────────────────────│
    │valid1            ││    ┌───┐                                                       │
    │                  ││────┘   └───────────────────────────────────────────────────────│
    │valid10           ││                                        ┌───┐                   │
    │                  ││────────────────────────────────────────┘   └───────────────────│
    │valid11           ││                                            ┌───┐               │
    │                  ││────────────────────────────────────────────┘   └───────────────│
    │valid12           ││                                                ┌───┐           │
    │                  ││────────────────────────────────────────────────┘   └───────────│
    │valid13           ││                                                    ┌───┐       │
    │                  ││────────────────────────────────────────────────────┘   └───────│
    │valid2            ││        ┌───┐                                                   │
    │                  ││────────┘   └───────────────────────────────────────────────────│
    │valid3            ││            ┌───┐                                               │
    │                  ││────────────┘   └───────────────────────────────────────────────│
    │valid4            ││                ┌───┐                                           │
    │                  ││────────────────┘   └───────────────────────────────────────────│
    │valid5            ││                    ┌───┐                                       │
    │                  ││────────────────────┘   └───────────────────────────────────────│
    │valid6            ││                        ┌───┐                                   │
    │                  ││────────────────────────┘   └───────────────────────────────────│
    │valid7            ││                            ┌───┐                               │
    │                  ││────────────────────────────┘   └───────────────────────────────│
    │valid8            ││                                ┌───┐                           │
    │                  ││────────────────────────────────┘   └───────────────────────────│
    │valid9            ││                                    ┌───┐                       │
    │                  ││────────────────────────────────────┘   └───────────────────────│
    │valid             ││                    ┌───────────────────────────────────────────│
    │                  ││────────────────────┘                                           │
    │                  ││────────────────┬───┬───┬───┬───┬───┬───┬───┬───┬───┬───┬───┬───│
    │value             ││ 00             │0D │01 │02 │03 │04 │05 │06 │07 │08 │09 │0A │0B │
    │                  ││────────────────┴───┴───┴───┴───┴───┴───┴───┴───┴───┴───┴───┴───│
    └──────────────────┘└────────────────────────────────────────────────────────────────┘ |}]
;;

let random ~cycles num_data =
  let waves, sim, valid = init ~cycles num_data in
  Cyclesim.cycle sim;
  for i = 0 to num_data - 1 do
    Array.iter valid ~f:(fun v -> v := if Random.int 2 = 0 then Bits.vdd else Bits.gnd);
    Cyclesim.cycle sim;
    valid.(i) := Bits.gnd
  done;
  Array.iter valid ~f:(fun v -> v := Bits.gnd);
  for _ = 1 to cycles do
    Cyclesim.cycle sim
  done;
  Cyclesim.cycle sim;
  Waveform.print
    ~display_width:86
    ~display_height:(11 + (num_data * 2))
    ~wave_width:1
    waves
;;

let%expect_test "random valids" =
  random ~cycles:2 7;
  (* Note; [07] is shown when nothing is valid (no inputs are high - [valid] out will be
     low).  Otherwise by insepction this looks correct. *)
  [%expect
    {|
    ((cycle 0) (reductions (3 3 1)))
    ((cycle 1) (reductions (3)))
    ┌Signals───────────┐┌Waves───────────────────────────────────────────────────────────┐
    │clock             ││┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ ┌─┐ │
    │                  ││  └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─┘ └─│
    │clear             ││                                                                │
    │                  ││────────────────────────────────────────────                    │
    │valid1            ││    ┌───┐           ┌───┐                                       │
    │                  ││────┘   └───────────┘   └───────────────────                    │
    │valid2            ││    ┌───┐               ┌───────┐                               │
    │                  ││────┘   └───────────────┘       └───────────                    │
    │valid3            ││    ┌───────┐               ┌───┐                               │
    │                  ││────┘       └───────────────┘   └───────────                    │
    │valid4            ││    ┌───────┐           ┌───┐                                   │
    │                  ││────┘       └───────────┘   └───────────────                    │
    │valid5            ││            ┌───┐   ┌───┐                                       │
    │                  ││────────────┘   └───┘   └───────────────────                    │
    │valid6            ││        ┌───┐       ┌───────┐                                   │
    │                  ││────────┘   └───────┘       └───────────────                    │
    │valid7            ││    ┌───────────┐       ┌───────┐                               │
    │                  ││────┘           └───────┘       └───────────                    │
    │valid             ││            ┌───────────┐   ┌───────────┐                       │
    │                  ││────────────┘           └───┘           └───                    │
    │                  ││────────┬───┬───┬───┬───┬───┬───┬───────┬───                    │
    │value             ││ 00     │07 │01 │03 │05 │07 │01 │02     │07                     │
    │                  ││────────┴───┴───┴───┴───┴───┴───┴───────┴───                    │
    └──────────────────┘└────────────────────────────────────────────────────────────────┘ |}]
;;
